from batou.component import Component
from batou import UpdateNeeded
import os.path
import shutil


class Clone(Component):

    # BBB the namevar used to be target, but all other VCS components have url
    namevar = 'url_or_target'
    url = None
    target = '.'
    source = None  # BBB

    update_unpinned = False

    def configure(self):
        if self.source is None:
            self.url = self.url_or_target
        else:
            self.url = self.source
            self.target = self.url_or_target

        self.git_dir = '{}/.git'.format(self.target)

    def verify(self):
        if not os.path.exists(self.git_dir):
            raise UpdateNeeded()
        if self.update_unpinned:
            raise UpdateNeeded()

    def update(self):
        if (os.path.exists(self.target) and not
                os.path.isdir(self.git_dir)):
            # Clean any non-git residuals
            shutil.rmtree(self.target)
        if not os.path.exists(self.target):
            self.cmd('git clone {0} {1}'.format(self.url, self.target))
        with self.chdir(self.target):
            self.cmd('git pull')
            self.cmd('git submodule update --init --recursive')



from batou.environment import Environment
from batou.remote import RemoteDeployment
from batou.utils import cmd
import pytest


@pytest.mark.slow
def test_remote_deployment_initializable(sample_service):
    cmd('hg init')
    with open('.hg/hgrc', 'w') as f:
        f.write('[paths]\ndefault=https://example.com')
    env = Environment('test-with-env-config')
    env.load()
    env.configure()
    RemoteDeployment(env, False)



from batou import UpdateNeeded
from batou.component import Component
from batou.lib.file import Directory
import logging
import os.path
import re


logger = logging.getLogger(__name__)


class Clone(Component):

    namevar = 'url'
    target = '.'
    revision = None
    branch = None
    vcs_update = True

    _revision_pattern = re.compile('parent: \d+:([a-f0-9]+) ')

    def configure(self):
        assert self.revision_or_branch
        self.target = self.map(self.target)
        self += Directory(self.target)

    def verify(self):
        with self.chdir(self.target):
            if not os.path.exists('.hg'):
                raise UpdateNeeded()

            if not self.vcs_update:
                return

            if self.has_changes:
                logger.error(
                    'Hg clone at {} has changes, refused to update.'.format(
                        self.target))
                return
            if self.has_outgoing_changesets:
                logger.error('Hg clone at {} has outgoing changesets, '
                             'refused to update.'.format(self.target))
                return

            if self.revision and self.current_revision != self.revision:
                raise UpdateNeeded()
            if (self.branch and (
                    self.current_branch != self.branch or
                    self.has_incoming_changesets)):
                raise UpdateNeeded()

    @property
    def revision_or_branch(self):
        # Mercurial often takes either a revision or a branch.
        return self.revision or self.branch

    @property
    def current_revision(self):
        stdout, stderr = self.cmd(
            self.expand('LANG=C hg --cwd {{component.target}} summary | '
                        'grep parent:'))
        match = self._revision_pattern.search(stdout)
        if not match:
            return None
        return match.group(1)

    @property
    def current_branch(self):
        stdout, stderr = self.cmd('hg branch')
        return stdout.strip()

    @property
    def has_incoming_changesets(self):
        try:
            self.cmd('hg incoming -q -l1', silent=True)
        except RuntimeError, e:
            returncode = e.args[1]
            if returncode == 1:
                return False
            raise
        return True

    @property
    def has_outgoing_changesets(self):
        try:
            with self.chdir(self.target):
                self.cmd('hg outgoing -q -l1', silent=True)
        except RuntimeError, e:
            returncode = e.args[1]
            if returncode == 1:
                return False
            raise
        return True

    @property
    def has_changes(self):
        with self.chdir(self.target):
            stdout, stderr = self.cmd('hg status -q')
        return bool(stdout.strip())

    def update(self):
        with self.chdir(self.target):
            if not os.path.exists('.hg'):
                self.cmd(self.expand(
                    'hg clone -u {{component.revision_or_branch}} '
                    '{{component.url}} .'))
            else:
                self.cmd('hg pull')
                self.cmd(self.expand(
                    'hg up --clean {{component.revision_or_branch}}'))

    def last_updated(self):
        with self.chdir(self.target):
            if not os.path.exists('.hg'):
                return None
            stdout, stderr = self.cmd(
                'hg log -r %s --template "{date|hgdate}\n"' %
                self.current_revision)
            timestamp, offset = stdout.split()
            return float(timestamp) - float(offset)



from ..cron import CronJob, CronTab
import batou.vfs
import os.path


def test_collects_cronjobs_into_crontab(root):
    root.environment.vfs_sandbox = batou.vfs.Developer(root.environment, None)
    root.component += CronJob('command1', timing='* * * * *')
    root.component += CronJob('command2', timing='* * * * *')
    root.component += CronTab()
    root.component.deploy()
    crontab = open(os.path.join(
        root.environment.workdir_base, 'mycomponent/crontab')).read()
    assert 'command1' in crontab
    assert 'command2' in crontab



from batou.environment import Environment
from mock import Mock
import pytest


def test_environment_should_raise_if_no_config_file(tmpdir):
    e = Environment(u'foobar')
    with pytest.raises(ValueError):
        e.load()


def test_load_should_use_defaults(sample_service):
    e = Environment(u'test-without-env-config')
    e.load()
    assert None == e.host_domain
    assert 'default' == e.branch


def test_load_should_use_config(sample_service):
    e = Environment(u'test-with-env-config')
    e.load()
    assert e.service_user == u'joe'
    assert e.host_domain == u'example.com'
    assert e.branch == u'release'


def test_load_ignores_predefined_environment_settings(sample_service):
    e = Environment(u'test-with-env-config')
    e.service_user = u'bob'
    e.host_domain = u'sample.com'
    e.branch = u'default'
    e.load()
    assert e.service_user == u'bob'
    assert e.host_domain == u'sample.com'
    assert e.branch == u'default'


def test_load_sets_up_overrides(sample_service):
    e = Environment('test-with-overrides')
    e.load()
    assert e.overrides == {'hello1': {'asdf': '1'}}


def test_loads_configures_vfs_sandbox(sample_service):
    e = Environment('test-with-vfs-sandbox')
    e.load()
    assert e.vfs_sandbox.map('/asdf').endswith('work/_/asdf')
    assert e.map('/asdf').endswith('work/_/asdf')


def test_default_sandbox_is_identity():
    e = Environment(u'foo')
    assert e.map('/asdf') == '/asdf'


def test_get_host_raises_keyerror_if_unknown():
    e = Environment(u'name')
    with pytest.raises(KeyError):
        e.get_host('asdf')


def test_get_host_normalizes_hostname():
    e = Environment(u'name')
    e.hosts['asdf.example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('asdf')
    assert host == e.get_host('asdf.example.com')


def test_normalize_hostname_regression_11156():
    # The issue here was that we used "rstrip" which works on characters,
    # not substrings. Having the domain (example.com) start with an eee
    # causes the hostname to get stripped of it's "eee"s ending in an
    # empty hostname accidentally.
    e = Environment(u'name')
    e.hosts['eee.example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('eee')


def test_get_host_without_subdomain_also_works():
    e = Environment(u'name')
    e.hosts['example.com'] = host = Mock()
    e.host_domain = 'example.com'
    assert host == e.get_host('example.com')


def test_get_root_raises_keyerror_on_nonassigned_component():
    e = Environment(u'foo')
    with pytest.raises(KeyError):
        e.get_root('asdf', 'localhost')


def test_multiple_components():
    e = Environment(u'test-multiple-components')
    e.load()
    components = dict(
        (host, list(sorted(c.name for c in e.roots_in_order(host=host))))
        for host in sorted(e.hosts.keys()))
    assert components == dict(
        localhost=['component1', 'component2'],
        otherhost=['component3', 'component4'],
        thishost=['component5', 'component6'])



import logging
import os
import os.path
import subprocess
import traceback


# Satisfy flake8 and support testing.
try:
    channel
except NameError:
    channel = None

deployment = None


class Deployment(object):

    environment = None

    def __init__(self, environment, host):
        self.environment = environment
        self.host = host

    def deploy(self, root):
        root = self.environment.get_root(root, self.host)
        root.component.deploy()


def lock():
    # XXX implement!
    pass


def cmd(c):
    return subprocess.check_output(
        [c], shell=True)


def ensure_repository():
    target = target_directory()
    if not os.path.exists(target):
        os.mkdir(target)
    if not os.path.exists(target + '/.hg'):
        cmd("hg init {}".format(target))
    return target


def current_heads():
    target = target_directory()
    os.chdir(target)
    result = []
    id = cmd('hg id -i').strip()
    if id == '000000000000':
        return [id]
    heads = cmd('hg heads')
    for line in heads.split('\n'):
        if not line.startswith('changeset:'):
            continue
        line = line.split(':')
        result.append(line[2])
    return result


def pull_code(upstream):
    # TODO Make choice of VCS flexible
    target = target_directory()
    os.chdir(target)
    # Phase 1: update working copy
    # XXX manage certificates
    cmd("hg pull {}".format(upstream))


def unbundle_code():
    # TODO Make choice of VCS flexible
    # XXX does this protect us from accidental new heads?
    target = target_directory()
    os.chdir(target)
    cmd('hg -y unbundle batou-bundle.hg')


def update_working_copy(branch):
    cmd("hg up -C {}".format(branch))
    id = cmd("hg id -i").strip()
    return id


def build_batou(deployment_base):
    target = target_directory()
    os.chdir(os.path.join(target, deployment_base))
    cmd('./batou --help')


def setup_deployment(deployment_base, env_name, host_name, overrides):
    from batou.environment import Environment

    target = target_directory()
    os.chdir(os.path.join(target, deployment_base))

    environment = Environment(env_name)
    environment.load()
    environment.overrides = overrides
    environment.configure()

    global deployment
    deployment = Deployment(environment, host_name)


def deploy(root):
    deployment.deploy(root)


def target_directory():
    # XXX make configurable?
    return os.path.expanduser('~/deployment')


def setup_logging(loggers, level):
    ch = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s %(levelname)s: %(message)s')
    ch.setFormatter(formatter)
    for logger in loggers:
        logger = logging.getLogger(logger)
        logger.setLevel(level)
        logger.addHandler(ch)


def send_file(name):
    code = None
    with open(name, 'w') as f:
        while code != 'finish':
            code, data = channel.receive()
            f.write(data)
    return 'OK'


if __name__ == '__channelexec__':
    setup_logging(['batou'], logging.INFO)
    while not channel.isclosed():
        task, args, kw = channel.receive()
        try:
            result = locals()[task](*args, **kw)
        except Exception, e:
            tb = traceback.format_exc()
            result = ('batou-remote-core-error', tb)
        channel.send(result)



from batou.component import Component
from batou.environment import Environment
import os
import pytest


@pytest.fixture
def root(tmpdir):
    environment = Environment('test', str(tmpdir))
    environment._set_defaults()
    os.chdir(str(tmpdir))

    class MyComponent(Component):
        defdir = str(tmpdir)
    environment.components['mycomponent'] = MyComponent
    root = environment.add_root('mycomponent', 'host')
    root.prepare()
    root.component.deploy()
    return root


@pytest.fixture(scope='session', autouse=True)
def ensure_workingdir(request):
    working_dir = os.getcwd()

    def go_back():
        os.chdir(working_dir)
    request.addfinalizer(go_back)



from batou.component import Component
from batou.lib.file import File


class Tick(Component):

    def configure(self):
        tick = File('tick.sh',
                    source='tick.sh',
                    mode=0755)
        self += tick
        self.provide('programs',
                     dict(name='tick',
                          path=tick.path,
                          priority=10))



from batou.component import HookComponent, Component, platform
from batou.lib.file import File
import os.path
import pkg_resources


class RotatedLogfile(HookComponent):

    namevar = 'path'
    key = 'batou.lib.logrotate:RotatedLogfile'

    args = ''
    prerotate = None
    postrotate = None

    def configure(self):
        super(RotatedLogfile, self).configure()
        self.path = os.path.join(self.workdir, self.path)
        self.path = self.map(self.path)
        self.args = map(str.strip, self.args.split(','))


class Logrotate(Component):

    common_config = ''
    logrotate_template = pkg_resources.resource_string(
        __name__, 'resources/logrotate.in')

    def configure(self):
        self.logfiles = self.require(RotatedLogfile.key, host=self.host)

        config = self.common_config + self.logrotate_template
        self.logrotate_conf = File('logrotate.conf', content=config)
        self += self.logrotate_conf


@platform('gocept.net', Logrotate)
class GoceptNetRotatedLogrotate(Component):

    def configure(self):
        user = self.environment.service_user
        user_logrotate_conf = os.path.join('/var/spool/logrotate/', user)
        self += File(
            user_logrotate_conf,
            ensure='symlink',
            link_to=self.parent.logrotate_conf.path)



"""Manage service components for individual hosts."""


class Host(object):
    """Description of service components for a host."""

    def __init__(self, fqdn):
        self.fqdn = fqdn
        self.name = self.fqdn.split('.')[0]



from ..nagios import NagiosServer, Service


def test_server_template(root):
    service = Service(
        'http',
        command='http',
        args='-H localhost -u /login.html',
        depend_on=[('localhost', 'Supervisor')])
    root.component |= service
    server = NagiosServer()
    root.component += server
    assert """\
# Generated from template; don't edit manually!

#
# http
#
define service {
    use         generic-service
    host_name   host
    service_description http
    check_command http!-H localhost -u /login.html
    servicegroups direct
}

define servicedependency {
    use generic-servicedependency
    host_name localhost
    dependent_host_name host
    service_description Supervisor
    dependent_service_description http
}
""" == server.sub_components[-1].content



from ..buildout import Buildout
from ..file import File
import mock
import os.path
import pkg_resources
import pytest


def buildout(**kw):
    buildout = Buildout(**kw)
    buildout.cmd = mock.Mock()
    return buildout


def test_update_should_pass_config_file_name(root):
    b = buildout(
        python='2.7', setuptools='1.0',
        config=File('myown.cfg'))
    root.component += b
    b.update()

    assert b.cmd.call_count == 1
    calls = iter(x[1][0] for x in b.cmd.mock_calls)
    assert calls.next() == 'bin/buildout -t 3 -c "myown.cfg"'


def test_update_should_pass_custom_timeout(root):
    b = buildout(
        python='2.7', setuptools='1.0',
        timeout=40)
    b.update()

    assert b.cmd.call_count == 1
    calls = iter(x[1][0] for x in b.cmd.mock_calls)
    assert calls.next() == 'bin/buildout -t 40 -c "buildout.cfg"'


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_runs_buildout_successfully(root):
    b = Buildout(
        python='2.7', version='2.2', setuptools='1.0',
        config=File('buildout.cfg', source=pkg_resources.resource_filename(
            __name__, 'buildout-example.cfg')))
    root.component += b
    root.component.deploy()
    assert os.path.isdir(os.path.join(
        root.environment.workdir_base, 'mycomponent/parts'))
    assert os.path.isfile(os.path.join(
        root.environment.workdir_base, 'mycomponent/bin/py'))



# -*- coding: utf-8 -*-
from batou.lib.archive import Extract
from pkg_resources import resource_filename
import os
import pytest
import sys


def test_unknown_extension_raises():
    extract = Extract('example.unknown')
    extract.workdir = ''
    with pytest.raises(ValueError):
        extract.configure()


def test_untar_extracts_archive_to_target_directory(root):
    extract = Extract(
        resource_filename(__name__, 'example.tar.gz'),
        target='example')
    root.component += extract
    root.component.deploy()
    assert os.listdir(unicode(extract.target)) == [u'foo']


def test_untar_can_strip_paths_off_archived_files(root):
    extract = Extract(
        resource_filename(__name__, 'example.tar.gz'),
        target='example', strip=1)
    root.component += extract
    root.component.deploy()
    assert os.listdir(unicode(extract.target)) == [u'bar']


def test_zip_extracts_archive_to_target_directory(root):
    extract = Extract(
        resource_filename(__name__, 'example.zip'),
        target='example')
    root.component += extract
    root.component.deploy()
    assert os.listdir(unicode(extract.target)) == [u'foo']


@pytest.mark.slow
@pytest.mark.skipif(sys.platform != 'darwin', reason='only runs on OS X')
def test_dmg_extracts_archive_to_target_directory(root):
    extract = Extract(
        resource_filename(__name__, 'example.dmg'),
        target='example')
    root.component += extract
    root.component.deploy()

    assert os.listdir(unicode(extract.target)) == [
        u' ', u'a\u0308sdf.txt', u'example.app']

    # ' ' is a symlink which stays one after copying:
    assert os.path.islink(extract.target + '/ ')
    start_bin = extract.target + '/example.app/MacOS/start.bin'
    with open(start_bin) as start_bin:
        assert start_bin.read() == 'I start the example app! ;)'


def test_dmg_does_not_support_strip(root):
    extract = Extract(
        resource_filename(__name__, 'example.dmg'),
        strip=1,
        target='example')
    with pytest.raises(ValueError) as e:
        root.component += extract
        assert e.value.args[0] == 'Strip is not supported by DMGExtractor'



from batou.component import Component, RootComponent, platform
from batou import UpdateNeeded
from mock import Mock
import batou
import os
import os.path
import pytest


class SampleComponent(Component):
    updated = False
    needs_update = True

    def verify(self):
        if self.needs_update:
            raise batou.UpdateNeeded()

    def update(self):
        self.updated = True


def test_init_with_no_arguments_creates_plain_component():
    component = Component()
    # Lazy initialized attribute
    assert not hasattr(component, 'sub_components')


def test_plain_component_runs_noop_configure_verify_update():
    component = Component()
    component.configure()
    component.verify()
    component.update()


def test_init_component_with_namevar_uses_first_argument():
    class SampleComponent(Component):
        namevar = 'asdf'
    component = SampleComponent('foobar')
    assert 'foobar' == component.asdf


def test_init_component_with_namevar_fails_without_argument():
    class SampleComponent(Component):
        namevar = 'asdf'
    with pytest.raises(ValueError):
        SampleComponent()


def test_init_keyword_args_update_dict():
    component = Component(foobar=1)
    assert 1 == component.foobar


def test_prepare_sets_up_vars(root):
    assert root.component.environment is root.environment
    assert root.component.host is root.host
    assert root.component.root is root
    assert root.component.parent is None


def test_op_orassignment_ignores_none(root):
    root.component |= None


def test_recursive_sub_component_iterator(root):
    for x in range(3):
        c = Component(name='x{}'.format(x))
        root.component += c
        for y in range(2):
            c2 = Component(name='x{}y{}'.format(x, y))
            c += c2

    recursive = list(x.name for x in root.component.recursive_sub_components)
    assert (['x0', 'x0y0', 'x0y1',
             'x1', 'x1y0', 'x1y1',
             'x2', 'x2y0', 'x2y1'] == recursive)


def test_op_orassignment_ignores_already_preapred_component(root):
    class Counting(Component):
        x = 1

        def configure(self):
            self.x += 1
    c = Counting()
    root.component |= c
    assert c.x == 2
    root.component |= c
    assert c.x == 2


def test_prepare_calls_configure():
    component = Component()
    component.configure = Mock()
    component.prepare(Mock(), Mock())
    assert component.configure.called


def test_prepare_configures_applicable_platforms_as_subcomponents(root):
    class MyComponent(Component):
        pass

    class MyOtherComponent(MyComponent):
        pass

    @platform('testplatform', MyComponent)
    class MyPlatform(Component):
        pass

    @platform('otherplatform', MyComponent)
    class MyOtherPlatform(Component):
        pass

    root.environment.platform = 'testplatform'
    component = MyComponent()
    root.component += component
    assert 1 == len(component.sub_components)
    assert isinstance(component.sub_components[0], MyPlatform)
    # Because we currently have no defined behaviour about this
    # I'm playing this rather safe: a sub-class of a component does not
    # automatically get to have the same platform components applied.
    other_component = MyOtherComponent()
    root.component += other_component
    assert 0 == len(other_component.sub_components)


def test_deploy_empty_component_runs_without_error(root):
    root.component.deploy()


def test_deploy_update_performed_if_needed(root):
    component = SampleComponent(needs_update=True)
    root.component += component
    root.component.deploy()
    assert component.updated


def test_deploy_update_not_performed_if_not_needed(root):
    component = SampleComponent(needs_update=False)
    root.component += component
    root.component.deploy()
    assert not component.updated


def test_sub_components_are_deployed_first(root):
    log = []

    class MyComponent(Component):
        namevar = 'id'

        def verify(self):
            log.append('{}:verify'.format(self.id))
            raise batou.UpdateNeeded()

        def update(self):
            log.append('{}:update'.format(self.id))

    c1 = MyComponent('1')
    root.component += c1
    c1 += MyComponent('2')
    root.component.deploy()
    assert [u'2:verify', u'2:update', u'1:verify', u'1:update'] == log


def test_adding_subcomponents_configures_them_immediately(root):
    class MyComponent(Component):
        configured = False

        def configure(self):
            self.configured = True
    my = MyComponent()
    root.component += my
    assert my.configured


# AFIC = assert file is current
def test_afic_raises_if_nonexisting_file(root):
    component = Component()
    root.component += component
    with pytest.raises(batou.UpdateNeeded):
        component.assert_file_is_current('idonotexist')


def test_afic_doesnt_raise_if_file_exists_but_no_reference_is_given(root):
    component = Component()
    root.component += component
    component.assert_file_is_current(__file__)


def test_afic_raises_if_file_isolder_than_reference(tmpdir, root):
    component = Component()
    root.component += component
    with pytest.raises(batou.UpdateNeeded):
        component.assert_file_is_current(__file__, [str(tmpdir)])


# ANSC = assert no subcomponent changes

def test_ansc_raises_if_subcomponent_changed(root):
    c2 = Component()
    root.component += c2
    c2.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_subcomponent_changes()


def test_ansc_does_not_raise_if_no_subcomponent_changed(root):
    c2 = Component()
    root.component += c2
    c2.changed = False
    root.component.assert_no_subcomponent_changes()


# ACIC = assert component is current

def test_acic_raises_if_no_reference(root):
    c = Component()
    c.last_updated = Mock(return_value=None)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current(c2)


def test_acic_raises_if_older_reference(root):
    c = Component()
    c.last_updated = Mock(return_value=20)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current(c2)


def test_acic_does_not_raise_if_current(root):
    c = Component()
    c.last_updated = Mock(return_value=21)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    c.assert_component_is_current(c2)


def test_acic_does_not_raise_if_newer(root):
    c = Component()
    c.last_updated = Mock(return_value=22)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    c.assert_component_is_current(c2)


def test_acic_accepts_multiple_components(root):
    c = Component()
    c.last_updated = Mock(return_value=20)
    root.component += c
    c2 = Component()
    c2.last_updated = Mock(return_value=21)
    with pytest.raises(batou.UpdateNeeded):
        c.assert_component_is_current([c2, c2])


def test_cmd_returns_output():
    c = Component()
    assert ('1\n', '') == c.cmd('echo 1')


def test_cmd_raises_if_error():
    c = Component()
    with pytest.raises(RuntimeError):
        c.cmd('non-existing-command')


def test_cmd_returns_output_if_ignore_returncode():
    c = Component()
    out, err = c.cmd('echo important output && false', silent=True,
                     ignore_returncode=True)
    assert 'important output\n' == out


def test_touch_creates_new_file(tmpdir):
    reference = str(tmpdir / 'reference')
    assert not os.path.exists(reference)
    c = Component()
    c.touch(reference)
    assert os.path.exists(reference)


def test_touch_updates_mtime_leaves_content_intact(tmpdir):
    reference = str(tmpdir / 'reference')
    with open(reference, 'w') as r:
        r.write('Hello world')
    c = Component()
    os.utime(reference, (0, 0))
    c.touch(reference)
    assert os.stat(reference).st_mtime != 0
    with open(reference, 'r') as r:
        assert 'Hello world' == r.read()


def test_expand(root):
    assert root.component.expand('Hello {{host.fqdn}}') == (
        'Hello host')


def test_templates(root):
    with open('sample', 'w') as template:
        template.write('Hello {{host.fqdn}}')
    assert root.component.template('sample') == (
        'Hello host\n')


def test_chdir_contextmanager_is_stackable():
    outer = os.getcwd()
    inner1 = os.path.join(os.path.dirname(__file__), 'fixture')
    inner2 = os.path.join(os.path.dirname(__file__))
    c = Component()
    with c.chdir(inner1):
        assert inner1 == os.getcwd()
        with c.chdir(inner2):
            assert inner2 == os.getcwd()
        assert inner1 == os.getcwd()
    assert outer == os.getcwd()


def test_root_and_component_know_working_dir(root):
    assert root.workdir.endswith('/work/mycomponent')
    assert root.component.workdir.endswith('/work/mycomponent')


def test_root_component_repr():
    host = Mock()
    root = RootComponent('haproxy', object, host, [])
    assert repr(root).startswith('<RootComponent "haproxy" object at ')


def test_component_manages_working_dir(root):
    previous_workdir = os.getcwd()
    os.rmdir(root.workdir)

    class RememberWorkDir(Component):
        def verify(self):
            self.parent.i_was_in = os.getcwd()
    root.component += RememberWorkDir()
    root.component.deploy()
    assert os.path.isdir(root.workdir)
    assert (os.path.realpath(root.component.i_was_in) ==
            os.path.realpath(root.workdir))
    cwd = os.getcwd()
    # Need to ensure we resolve symlinks: on OS x /var/tmp may lead to
    # /private/...
    assert os.path.realpath(cwd) == os.path.realpath(previous_workdir)


def test_cmd_execution_failed_gives_command_in_exception():
    c = Component()
    with pytest.raises(RuntimeError) as e:
        c.cmd('asdf')
    assert str(e.value[0]) == 'Command "asdf" returned unsuccessfully.'
    assert e.value[1] == 127


def test_cmd_should_not_stop_if_process_expects_input():
    c = Component()
    stdout, stderr = c.cmd('cat')
    # The assertion is, that the test doesn't get stuck .


def test_last_updated_not_implemented_on_base():
    c = Component()
    with pytest.raises(NotImplementedError):
        c.last_updated()


def test_require_one_convenience_api_returns_scalar():
    c = Component()
    c.require = Mock(return_value=[1])
    assert 1 == c.require_one('asdf')


def test_require_one_convenience_raises_if_no_result():
    c = Component()
    c.require = Mock(return_value=[])
    with pytest.raises(KeyError):
        c.require_one('asdf')


def test_require_one_convenience_raises_if_more_results():
    c = Component()
    c.require = Mock(return_value=[1, 2])
    with pytest.raises(KeyError):
        c.require_one('asdf')


def test_assert_cmd_when_succesful():
    c = Component()
    c.assert_cmd('true')


def test_assert_cmd_when_unsuccessful():
    c = Component()
    with pytest.raises(UpdateNeeded):
        c.assert_cmd('false')


def test_assert_no_changes_local_does_not_raise(root):
    root.component.assert_no_changes()


def test_assert_no_changes_local_raises(root):
    root.component.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_changes()


def test_assert_no_changes_recursive_does_not_raise(root):
    root.component += Component()
    root.component.assert_no_changes()


def test_assert_no_changes_recursive_raises(root):
    c2 = Component()
    root.component += c2
    c2.changed = True
    with pytest.raises(batou.UpdateNeeded):
        root.component.assert_no_changes()


def test_root_overrides_missing_attribute_raises(root):
    root.environment.overrides = {'mycomponent': {'asdf': 1}}
    with pytest.raises(KeyError):
        root.prepare()


def test_root_overrides_existing_attribute(root):
    class OtherComponent(Component):
        asdf = None
    root.environment.components['mycomponent'] = OtherComponent
    root.environment.overrides = {'mycomponent': {'asdf': 1}}
    root.prepare()
    assert root.component.asdf == 1



from batou.component import Component
from batou.lib.python import VirtualEnv, Package
from batou.update import update_bootstrap
from batou.utils import cmd
import os
import os.path
import pytest


@pytest.mark.slow
@pytest.mark.timeout(240)
def test_package_venv_installations(root):
    # this is a really nasty test ...
    base_dir = os.path.join(
        os.path.dirname(__file__), '../../../../')

    # copy the "venv" example
    cmd('cp -a {}/examples/venvs/* .'.format(base_dir))
    cmd('rm -rf work')

    # ensure the dev link is OK
    update_bootstrap('', base_dir)

    # run batou, hope for the best. ;)
    stdout, stderr = cmd('./batou local dev localhost')
    for line in filter(None, stdout.split('\n')):
        assert line.startswith('Updating Py')
    assert "Updating Py24 >" in stdout
    assert "Updating Py25 >" in stdout
    assert "Updating Py26 >" in stdout
    assert "Updating Py27 >" in stdout
    assert "Updating Py31 >" in stdout
    assert "Updating Py32 >" in stdout
    assert "Updating Py33 >" in stdout
    assert stderr == ""


@pytest.mark.slow
@pytest.mark.timeout(60)
def test_updates_old_distribute_to_setuptools(root):
    class Playground(Component):
        def configure(self):
            self.venv = VirtualEnv('2.7')
            self += self.venv

    distribute = Package('distribute', version='0.6.34', timeout=10)
    setuptools = Package('setuptools', version='0.9.8', timeout=10)

    playground = Playground()
    root.component += playground
    playground.venv += distribute
    playground.deploy()

    playground.venv.sub_components.remove(distribute)
    playground.venv += setuptools
    playground.deploy()

    # We can't simply call verify, since distribute *still* manages to
    # manipulate the installation process and update itself, so our version
    # number is ignored and we get the most recent setuptools. *le major sigh*
    assert 'distribute-' not in playground.cmd(
        playground.workdir + '/bin/python -c '
        '"import setuptools; print setuptools.__file__"')[0]



from batou.component import Component


class Service(Component):
    """A generic component to provide a system service.

    Platform-specific components need to perform the work necessary
    to ensure startup and shutdown of the executable correctly.
    """

    namevar = 'executable'

    pidfile = None  # The pidfile as written by the services' executable.



from collections import defaultdict
import contextlib
import fcntl
import hashlib
import itertools
import os
import socket
import subprocess
import sys
import time
import logging


logger = logging.getLogger(__name__)


class MultiFile(object):

    def __init__(self, files):
        self.files = files

    def write(self, value):
        for file in self.files:
            file.write(value)

    def flush(self):
        for file in self.files:
            file.flush()


@contextlib.contextmanager
def locked(filename):
    # XXX can we make this not leave files around?
    with open(filename, 'a+') as lockfile:
        try:
            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            print >> sys.stderr, 'Could not acquire lock {}'.format(filename)
            raise RuntimeError(
                'cannot create lock "%s": more than one instance running '
                'concurrently?' % lockfile, lockfile)
        # publishing the process id comes handy for debugging
        lockfile.seek(0)
        lockfile.truncate()
        print >> lockfile, os.getpid()
        lockfile.flush()
        yield
        lockfile.seek(0)
        lockfile.truncate()


def flatten(listOfLists):
    return list(itertools.chain.from_iterable(listOfLists))


def notify(title, description):
    try:
        subprocess.check_call(['notify-send', title, description])
    except OSError:
        pass


def resolve(address):
    if ':' in address:
        host, port = address.split(':')
    else:
        host, port = address, None
    try:
        address = socket.gethostbyname(host)
    except socket.gaierror as e:
        raise socket.gaierror('%s (%s)' % (str(e), host))
    if port:
        address += ':%s' % port
    return address


class Address(object):
    """An internet service address that can be listened and connected to.

    The constructor address is expected to be the address that can be
    connected to. The listen address will be computed automatically.

    """

    def __init__(self, connect_address, port=None):
        if ':' in connect_address:
            connect, port = connect_address.split(':')
        else:
            connect = connect_address
        if port is None:
            raise ValueError('Need port for service address.')
        self.connect = NetLoc(connect, str(port))
        self.listen = NetLoc(resolve(connect), str(port))

    def __str__(self):
        return str(self.connect)


class NetLoc(object):
    """A network location specified by host and port."""

    def __init__(self, host, port=None):
        self.host = host
        self.port = port

    def __str__(self):
        result = self.host
        if self.port:
            result += ':' + self.port
        return result


def revert_graph(graph):
    graph = ensure_graph_data(graph)
    reverse_graph = defaultdict(set)
    for node, dependencies in graph.items():
        # Ensure all nodes will exist
        reverse_graph[node]
        for dependency in dependencies:
            reverse_graph[dependency].add(node)
    return reverse_graph


def ensure_graph_data(graph):
    # Ensure that all nodes exist as keys even if they don't have outgoing
    # relations.
    for node, relations in list(graph.items()):
        for relation in relations:
            if relation not in graph:
                graph[relation] = set()
    return graph


class CycleError(ValueError):
    """Graph contains at least one cycle."""


def remove_nodes_without_outgoing_edges(graph):
    for node, dependencies in list(graph.items()):
        if not dependencies:
            del graph[node]


def topological_sort(graph):
    """Take a directed graph and provide a topological sort of all nodes.

    The graph is given as

    {node: [dependency, dependency], ...}

    If the graph has cycles a ValueError will be raised.
    """
    graph = ensure_graph_data(graph)
    sorted = []
    reverse_graph = revert_graph(graph)
    roots = [node for node, incoming in reverse_graph.items()
             if not incoming]
    while roots:
        root = roots.pop()
        sorted.append(root)
        for node in list(graph[root]):
            graph[root].remove(node)
            reverse_graph[node].remove(root)
            if not reverse_graph[node]:
                roots.append(node)
    if any(graph.values()):
        # Simplify the graph a bit to make it easier to spot the cycle.
        remove_nodes_without_outgoing_edges(graph)
        raise CycleError(dict(graph))
    return sorted


class CmdExecutionError(RuntimeError):
    pass


def cmd(cmd, silent=False, ignore_returncode=False, communicate=True,
        env=None, acceptable_returncodes=[0]):
    if not isinstance(cmd, basestring):
        # We use `shell=True`, so the command needs to be a single string and
        # we need to pay attention to shell quoting.
        quoted_args = []
        for arg in cmd:
            arg = arg.replace('\'', '\\\'')
            if ' ' in arg:
                arg = "'{}'".format(arg)
            quoted_args.append(arg)
        cmd = ' '.join(quoted_args)
    if env is not None:
        add_to_env = env
        env = os.environ.copy()
        env.update(add_to_env)
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        shell=True,
        env=env)
    if not communicate:
        # XXX See #12550
        return process
    stdout, stderr = process.communicate()
    if process.returncode not in acceptable_returncodes:
        if not silent:
            print("$ {}".format(cmd))
            print("STDOUT")
            print("=" * 72)
            print(stdout)
            print("STDERR")
            print("=" * 72)
            print(stderr)
        if not ignore_returncode:
            raise CmdExecutionError(
                'Command "{}" returned unsuccessfully.'.format(cmd),
                process.returncode, stdout, stderr)
    return stdout, stderr


class Timer(object):

    def __init__(self, note):
        self.duration = 0
        self.note = note

    def __enter__(self):
        self.started = time.time()

    def __exit__(self, exc1, exc2, exc3):
        self.duration = time.time() - self.started
        logger.debug(self.note + ' took %fs' % self.duration)


def hash(path, function='md5'):
    h = getattr(hashlib, function)()
    for line in open(path):
        h.update(line)
    return h.hexdigest()



from batou.component import Component
from batou.environment import Environment
import os
import os.path
import pwd
import pytest


def test_parse_nonexisting_file_raises_ioerror(tmpdir):
    env = Environment('test', basedir=str(tmpdir))
    with pytest.raises(ValueError):
        env.load()


@pytest.fixture
def env():
    environment = Environment(
        'dev', os.path.dirname(__file__) + '/fixture/basic_service')
    environment.load()
    return environment


@pytest.fixture
def env_prod():
    environment = Environment(
        'production', os.path.dirname(__file__) + '/fixture/basic_service')
    environment.load()
    return environment


def test_service_options_are_set(env):
    assert env.base_dir.endswith('/fixture/basic_service')


def test_components_are_fully_loaded(env):
    assert sorted(env.components) == ['zeo', 'zope']


def test_production_environment_is_loaded(env_prod):
    assert env_prod.service_user == 'alice'
    assert env_prod.branch == 'production'
    assert env_prod.host_domain == 'example.com'

    assert sorted(env_prod.hosts) == [
        'host1.example.com', 'host2.example.com', 'host3.example.com']

    host1 = env_prod.hosts['host1.example.com']
    assert host1.name == 'host1'
    assert host1.fqdn == 'host1.example.com'
    host1_components = [x.name for x in env_prod.root_components
                        if x.host is host1]
    assert host1_components == ['zope']


def test_dev_environment_is_loaded(env):
    assert pwd.getpwuid(os.getuid()).pw_name == env.service_user
    assert env.branch == 'default'
    assert env.host_domain is None
    assert set(env.hosts.keys()) == set(['localhost', 'host2'])

    localhost = env.hosts['localhost']
    assert localhost.name == 'localhost'
    assert localhost.fqdn == 'localhost'
    root_components = set(
        [x.name for x in env.root_components if x.host is localhost])
    assert root_components == set(['zeo', 'zope'])

    zeo = env.get_root('zeo', 'localhost')
    zeo.prepare()
    assert isinstance(zeo.component, Component)
    assert zeo.name == 'zeo'


def test_component_has_features_set(env):
    env.configure()
    # The localhost ZEO doesn't specify features and thus gets all set
    zeo1 = env.get_root('zeo', 'localhost').component
    assert zeo1.features == ['test', 'test2']

    # ZEO on host2 has selected only 'test'
    zeo2 = env.get_root('zeo', 'host2').component
    assert zeo2.features == ['test']


def test_load_environment_with_overrides(env):
    env.overrides['zeo'] = {'port': '9002'}
    env.configure()
    zeo = env.get_root('zeo', 'localhost').component
    assert zeo.port == '9002'



import batou
import batou.c
import batou.template
import batou.utils
import contextlib
import logging
import os
import os.path
import sys
import types

logger = logging.getLogger(__name__)


def platform(name, component):
    """Class decorator to register a component class as a platform-component
    for the given platform and component.
    """
    def register_platform(cls):
        component.add_platform(name, cls)
        return cls
    return register_platform


def load_components_from_file(filename):
    # XXX protect against loading the same component name multiple times.
    components = {}

    defdir = os.path.dirname(filename)
    module_name = os.path.basename(defdir)
    module_path = 'batou.c.{}'.format(module_name)
    module = types.ModuleType(
        module_name, 'Component definition module for {}'.format(filename))
    sys.modules[module_path] = module
    setattr(batou.c, module_name, module)
    execfile(filename, module.__dict__)

    for candidate in module.__dict__.values():
        if candidate in [Component]:
            # Ignore anything we pushed into the globals before execution
            continue
        if not (isinstance(candidate, type) and
                issubclass(candidate, Component)):
            continue
        candidate.defdir = defdir
        components[candidate.__name__.lower()] = candidate

    return components


class Component(object):

    namevar = ''

    workdir = None

    changed = False
    _prepared = False

    def __init__(self, namevar=None, **kw):
        if self.namevar:
            if namevar is None:
                raise ValueError('Namevar %s required' % self.namevar)
            kw[self.namevar] = namevar
        self.__dict__.update(kw)

    @property
    def host(self):
        return self.root.host

    @property
    def environment(self):
        return self.root.environment

    # Configuration phase

    def prepare(self, root, parent=None):
        # XXX why can parent be None? Can the root be the last parent?
        # Can't the direct access to root be removed here?
        self.root = root
        self.parent = parent
        if parent:
            self.workdir = parent.workdir
        else:
            self.workdir = root.workdir
        self.sub_components = []
        self.configure()
        self += self.get_platform()
        self._prepared = True

    def configure(self):
        """Configure the component.

        At this point the component has been embedded into the overall
        structure, so that environment, host, root, and parent are set.

        Also, any environment-specific attributes have been set.

        At this point the component should add sub-components as required and
        use the configured values so that calls to verify/update can, without
        much overhead, perform their work.

        Also, any computation that happens in the configure phase is run while
        building the model so that this can be used to ensure some level of
        internal consistency before invoking the remote control layer.
        """
        pass

    # Deployment phase

    # We're context managers - to allow components to do setup/teardown nicely.

    def __enter__(self):
        pass

    def __exit__(self, type, value, tb):
        pass

    def deploy(self):
        # Remember: this is a tight loop - we need to keep this code fast.

        # Reset changed flag here to support triggering deploy() multiple
        # times. This is mostly helpful for testing, but who know.
        self.changed = False
        for sub_component in self.sub_components:
            sub_component.deploy()
            if sub_component.changed:
                self.changed = True

        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)
        with self.chdir(self.workdir), self:
            try:
                with batou.utils.Timer(
                        '{} verify()'.format(self._breadcrumbs)):
                    self.verify()
            except batou.UpdateNeeded:
                logger.info('Updating {}'.format(self._breadcrumbs))
                self.update()
                self.changed = True

    def verify(self):
        """Verify whether this component has been deployed correctly or needs
        to be updated.

        Raises the UpdateNeeded exception if an update is needed.

        Implemented by specific components.
        """
        pass

    def update(self):
        """Is called if the verify method indicated that an update is needed.

        This method needs to handle all cases to move the current deployed
        state of the component, whatever it is, to the desired new state.

        Implemented by specific components.

        """
        pass

    def last_updated(self):
        """An optional helper to indicate to other components a timestamp how
        new any changes in the target system related to this component are.

        Can be used to determine file ages, etc.
        """
        raise NotImplementedError

    # Sub-component mechanics

    def __add__(self, component):
        """Add a new sub-component.

        This will also automatically prepare the added component if it hasn't
        been prepared yet. Could have been prepared if it was configured in the
        context of a different component.

        """
        if component is not None:
            # Allow `None` components to flow right through. This makes the API
            # a bit more convenient in some cases, e.g. with platform handling.
            self.sub_components.append(component)
            self |= component
            component.parent = self
        return self

    def __or__(self, component):
        """Prepare a component in the context of this component but do not add
        it to the sub components.

        This allows executing 'configure' in the context of this component

        """
        if component is not None and not component._prepared:
            component.prepare(self.root, self)
        return self

    @property
    def recursive_sub_components(self):
        for sub in self.sub_components:
            yield sub
            for rec_sub in sub.recursive_sub_components:
                yield rec_sub

    # Platform mechanics

    @classmethod
    def add_platform(cls, name, platform):
        if not '_platforms' in cls.__dict__:
            cls._platforms = {}
        cls._platforms[name] = platform

    def get_platform(self):
        """Return the platform component for this component if one exists."""
        platforms = self.__class__.__dict__.get('_platforms', {})
        return platforms.get(self.environment.platform, lambda: None)()

    # Resource provider/user API

    def provide(self, key, value):
        self.environment.resources.provide(self.root, key, value)

    def require(self, key, host=None, strict=True, reverse=False):
        return self.environment.resources.require(
            self.root, key, host, strict, reverse)

    def require_one(self, key, host=None, strict=True, reverse=False):
        resources = self.require(key, host, strict, reverse)
        if len(resources) > 1:
            raise KeyError(
                'Expected only one result, got multiple for (key={}, host={})'.
                format(key, host))
        elif len(resources) == 0:
            raise KeyError(
                'Expected one result, got none for (key={}, host={})'.
                format(key, host))
        return resources[0]

    def assert_cmd(self, *args, **kw):
        try:
            kw['silent'] = True
            self.cmd(*args, **kw)
        except RuntimeError:
            raise batou.UpdateNeeded()

    def assert_file_is_current(self, reference, requirements=[], **kw):
        from batou.lib.file import File
        reference = File(reference, is_template=False)
        self |= reference
        reference.assert_component_is_current(
            [File(r, is_template=False) for r in requirements], **kw)

    def assert_component_is_current(self, requirements=[], **kw):
        if isinstance(requirements, Component):
            requirements = [requirements]
        reference = self.last_updated(**kw)
        if reference is None:
            logger.debug('assert_component_is_current({}, ...): No reference'.
                         format(self._breadcrumb))
            raise batou.UpdateNeeded()
        for requirement in requirements:
            self |= requirement
            required = requirement.last_updated(**kw)
            if reference < required:
                logger.debug('assert_component_is_current({}, {}): {} < {}'.
                             format(self._breadcrumb, requirement._breadcrumb,
                                    reference, required))
                raise batou.UpdateNeeded()

    def assert_no_subcomponent_changes(self):
        for component in self.sub_components:
            if component.changed:
                raise batou.UpdateNeeded()

    def assert_no_changes(self):
        if self.changed:
            raise batou.UpdateNeeded()
        self.assert_no_subcomponent_changes()

    def cmd(self, cmd, silent=False, ignore_returncode=False,
            communicate=True, env=None):
        return batou.utils.cmd(
            cmd, silent, ignore_returncode, communicate, env)

    def map(self, path):
        path = os.path.expanduser(path)
        if not path.startswith('/'):
            return os.path.normpath(os.path.join(self.workdir, path))
        return self.environment.map(path)

    def touch(self, filename):
        if os.path.exists(filename):
            os.utime(filename, None)
        else:
            open(filename, 'w').close()

    def expand(self, string, component=None, **kw):
        engine = batou.template.Jinja2Engine()
        args = self._template_args(component=component, **kw)
        return engine.expand(string, args, self._breadcrumbs)

    def template(self, filename, component=None):
        engine = batou.template.Jinja2Engine()
        return engine.template(
            filename, self._template_args(component=component))

    def _template_args(self, component=None, **kw):
        if component is None:
            component = self
        args = dict(
            host=self.host,
            environment=self.environment,
            component=component)
        args.update(kw)
        return args

    @contextlib.contextmanager
    def chdir(self, path):
        old = os.getcwd()
        try:
            os.chdir(path)
            yield
        finally:
            os.chdir(old)

    # internal methods

    @property
    def _breadcrumbs(self):
        result = ''
        if self.parent is not None:
            result += self.parent._breadcrumbs + ' > '
        result += self._breadcrumb
        return result

    @property
    def _breadcrumb(self):
        result = self.__class__.__name__
        name = self.namevar_for_breadcrumb
        if name:
            result += '({})'.format(name)
        return result

    @property
    def namevar_for_breadcrumb(self):
        return getattr(self, self.namevar, None)


class HookComponent(Component):
    """A component that provides itself as a resource."""

    def configure(self):
        self.provide(self.key, self)


class RootComponent(object):
    """Wrapper to manage top-level components assigned to hosts in an
    environment.

    Root components have a name and determine the initial working directory
    of the sub-components.

    """

    def __init__(self, name, environment, host, features):
        self.name = name
        self.environment = environment
        self.host = host
        self.features = features

    # XXX Should we turn this around and make the environment responsible for
    # putting those attributes on the root?

    @property
    def defdir(self):
        return self.environment.components[self.name].defdir

    @property
    def workdir(self):
        return os.path.join(
            self.environment.workdir_base, self.name)

    def prepare(self):
        self.component = self.environment.components[self.name]()
        if self.features:
            self.component.features = self.features
        self._overrides()
        self.component.prepare(self)

    def _overrides(self):
        if self.name not in self.environment.overrides:
            return
        for key, value in self.environment.overrides[self.name].items():
            if not hasattr(self.component, key):
                raise KeyError(
                    'Invalid override attribute "{}" for component {}'.format(
                        key, self.component))
            setattr(self.component, key, value)

    def __repr__(self):
        return '<%s "%s" object at %s>' % (
            self.__class__.__name__, self.name, id(self))



from batou.component import Component
from batou.lib.file import File
from batou.lib.python import VirtualEnv, Package
import contextlib
import os.path


class Buildout(Component):

    timeout = None
    extends = ()   # Extends need to be components that have a path
    use_default = True
    config = None
    additional_config = ()
    config_file_name = 'buildout.cfg'

    python = None
    executable = None
    distribute = None
    setuptools = None
    version = None

    build_env = {}  # XXX not frozen. :/

    def configure(self):
        if self.timeout is None:
            self.timeout = self.environment.timeout
        if self.use_default and not self.config:
            # We expect that your definition directory has a buildout.cfg
            self.config = File('buildout.cfg',
                               template_context=self.parent)
        if isinstance(self.config, File):
            self.config_file_name = self.config.path
        if isinstance(self.config, Component):
            self.config = [self.config]
        if not self.config:
            self.config = []
        self.config.extend(self.additional_config)
        for component in self.config:
            self += component

        venv = VirtualEnv(self.python, executable=self.executable)
        self += venv

        if not (self.distribute or self.setuptools):
            raise ValueError(
                'Either setuptools or distribute version must be specified')

        if self.distribute:
            venv += Package('distribute',
                            version=self.distribute,
                            check_package_is_module=False)

        if self.setuptools:
            venv += Package('setuptools',
                            version=self.setuptools)

        # Install without dependencies (that's just setuptools anyway), since
        # that could cause pip to pull in the latest version of setuptools,
        # regardless of the version we wanted to be installed above.
        venv += Package(
            'zc.buildout',
            version=self.version,
            dependencies=False)

    def verify(self):
        self.assert_file_is_current('bin/buildout')
        # XXX we can't be sure that all config objects are files!
        installed = File('.installed.cfg')
        self |= installed
        installed.assert_component_is_current(
            [File('bin/buildout')] + self.config)
        self.assert_file_is_current(
            '.batou.buildout.success', ['.installed.cfg'])
        self.assert_no_subcomponent_changes()

    def update(self):
        with safe_environment(self.build_env):
            self.cmd('bin/buildout -t {} -c "{}"'.format(
                self.timeout, self.config_file_name))
            self.touch('.batou.buildout.success')


@contextlib.contextmanager
def safe_environment(environment):
    old_env = os.environ.copy()
    for key, value in environment.items():
        old_env.setdefault(key, '')
        environment[key] = value.format(**old_env)
    try:
        os.environ.update(environment)
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_env)



from StringIO import StringIO
from batou.utils import hash
from batou.utils import resolve, MultiFile, locked, notify, Address
from batou.utils import revert_graph, topological_sort, flatten, NetLoc
from batou.utils import remove_nodes_without_outgoing_edges, cmd
import mock
import os
import pytest
import socket
import tempfile
import threading
import unittest


@mock.patch('socket.gethostbyname')
def test_host_without_port_resolves(ghbn):
    ghbn.return_value = '127.0.0.1'
    assert resolve('localhost') == '127.0.0.1'


@mock.patch('socket.gethostbyname')
def test_host_with_port_resolves_and_keeps_port(ghbn):
    ghbn.return_value = '127.0.0.1'
    assert resolve('localhost:8080') == '127.0.0.1:8080'


@mock.patch('socket.gethostbyname',
            side_effect=socket.gaierror('lookup failed'))
def test_socket_error_shows_hostname(ghbn):
    with pytest.raises(socket.gaierror) as e:
        resolve('localhost')
    assert str(e.value) == 'lookup failed (localhost)'


def test_flatten():
    assert [1, 2, 3, 4] == flatten([[1, 2], [3, 4]])


class MultiFileTests(unittest.TestCase):

    def test_write_and_flush_propation(self):
        file1 = StringIO()
        file2 = StringIO()
        multi = MultiFile([file1, file2])
        multi.write('asdf')
        multi.flush()
        self.assertEquals('asdf', file1.getvalue())
        self.assertEquals('asdf', file2.getvalue())


the_fake_lock = threading.Lock()


def fake_lock(lockfile, options):
    # Emulate fcntl's behaviour if we would be two processes.
    success = the_fake_lock.acquire(False)
    if not success:
        raise IOError


class LockfileContextManagerTests(unittest.TestCase):

    def setUp(self):
        self.files = []

    def tearDown(self):
        map(os.unlink, self.files)

    def tempfile(self):
        f = tempfile.mktemp()
        self.files.append(f)
        return f

    def test_lock_creates_file_and_writes_and_removes_pid(self):
        lockfile = self.tempfile()
        with locked(lockfile):
            pid = open(lockfile, 'r').read().strip()
            self.assertEquals(os.getpid(), int(pid))
        self.assertEquals('', open(lockfile, 'r').read())

    def test_lock_works_with_existing_file(self):
        lockfile = self.tempfile()
        f = open(lockfile, 'w')
        f.write('sadf')
        f.close()
        with locked(lockfile):
            pid = open(lockfile, 'r').read().strip()
            self.assertEquals(os.getpid(), int(pid))
        self.assertEquals('', open(lockfile, 'r').read())

    @mock.patch('fcntl.lockf', side_effect=fake_lock)
    def test_lock_cant_lock_twice(self, lockf):
        lockfile = self.tempfile()
        with locked(lockfile):
            locked2 = locked(lockfile)
            self.assertRaises(RuntimeError, locked2.__enter__)


class NotifyTests(unittest.TestCase):

    @mock.patch('subprocess.check_call')
    def test_notify_calls_notify_send(self, call):
        notify('foo', 'bar')
        call.assert_called_with(['notify-send', 'foo', 'bar'])

    @mock.patch('subprocess.check_call', side_effect=OSError)
    def test_notify_does_not_fail_if_os_call_fails(self, call):
        notify('foo', 'bar')


class AddressNetLocTests(unittest.TestCase):

    def test_address_without_implicit_or_explicit_port_fails(self):
        self.assertRaises(ValueError, Address, 'localhost')
        Address('localhost:8080')
        Address('localhost', 8080)

    def test_address_resolves_listen_address(self):
        address = Address('localhost:8080')
        self.assertEquals('127.0.0.1:8080', str(address.listen))
        self.assertEquals('localhost:8080', str(address.connect))

    def test_address_netloc_attributes(self):
        address = Address('localhost:8080')
        self.assertEquals('127.0.0.1', address.listen.host)
        self.assertEquals('8080', address.listen.port)
        self.assertEquals('localhost', address.connect.host)
        self.assertEquals('8080', address.connect.port)


def test_address_format_with_port():
    assert str(Address('127.0.0.1:8080').listen) == '127.0.0.1:8080'


def test_netloc_format_without_port():
    assert str(NetLoc('127.0.0.1')) == '127.0.0.1'


def test_revert_graph_no_edges_is_identical():
    graph = {1: set(), 2: set()}
    assert graph == dict(revert_graph(graph))


def test_revert_graph_one_edge_reverses():
    graph = {1: set([2])}
    assert {2: set([1]), 1: set()} == dict(revert_graph(graph))


def test_topological_sort_simple_chain():
    graph = {1: set([2]), 2: set([3]), 3: set([4])}
    assert [1, 2, 3, 4] == topological_sort(graph)


def test_topological_sort_multiple_paths():
    graph = {1: set([2, 3]), 2: set([3])}
    assert [1, 2, 3] == topological_sort(graph)


def test_topological_sort_raises_on_loop():
    graph = {1: set([2]), 2: set([3]), 3: set([1])}
    with pytest.raises(ValueError):
        topological_sort(graph)


def test_topological_sort_with_single_item():
    graph = {1: set()}
    assert [1] == topological_sort(graph)


def test_graph_remove_leafs():
    graph = {1: [1],
             2: []}
    remove_nodes_without_outgoing_edges(graph)
    assert graph == {1: [1]}


class Checksum(unittest.TestCase):

    fixture = os.path.join(
        os.path.dirname(__file__), 'fixture', 'component', 'haproxy.cfg')

    def test_hash_md5(self):
        self.assertEquals('ce0324fa445475e76182c0d114615c7b',
                          hash(self.fixture, 'md5'))

    def test_hash_sha1(self):
        self.assertEquals('164d8815aa839cca339e38054622b58ca80124a1',
                          hash(self.fixture, 'sha1'))


@mock.patch('subprocess.Popen')
def test_cmd_joins_list_args(popen):
    popen().communicate.return_value = ('', '')
    popen().returncode = 0
    cmd(['cat', 'foo', 'bar'])
    assert popen.call_args[0] == ('cat foo bar',)


@mock.patch('subprocess.Popen')
def test_cmd_quotes_spacey_args(popen):
    popen().communicate.return_value = ('', '')
    popen().returncode = 0
    cmd(['cat', 'foo', 'bar bz baz'])
    assert popen.call_args[0] == ("cat foo 'bar bz baz'",)
    cmd(['cat', 'foo', "bar 'bz baz"])
    assert popen.call_args[0] == (r"cat foo 'bar \'bz baz'",)


@mock.patch('subprocess.Popen')
def test_cmd_ignores_specified_returncodes(popen):
    popen.return_value.returncode = 4
    popen.return_value.communicate.return_value = '', ''
    with pytest.raises(RuntimeError):
        cmd('asdf')
    cmd('asdf', acceptable_returncodes=[0, 4])


@mock.patch('subprocess.Popen')
def test_cmd_returns_process_if_no_communicate(popen):
    process = mock.Mock()
    popen.return_value = process
    p = cmd(['asdf'], communicate=False)
    assert popen.communicate.call_count == 0
    assert p is process



from batou.component import Component
from batou.lib.file import File


class Hello(Component):

    magic_word = None
    other_word = None

    def configure(self):
        self += File('hello',
            content='The magic word is {{component.magic_word}}.\n'
                    'The other word is {{component.other_word}}.\n')



from batou.component import Component
from batou.lib.buildout import Buildout


class Py25(Component):

    def configure(self):
        buildout = Buildout(
            python='2.5',
            setuptools='1.3.2',
            version='1.7.1')
        self += buildout



from __future__ import print_function, unicode_literals
import ConfigParser
import StringIO
import fcntl
import os
import subprocess


NEW_FILE_TEMPLATE = """\
[batou]
members =
"""


class EncryptedConfigFile(object):
    """Wrap encrypted config files."""

    lockfd = None
    _cleartext = None

    # Additional GPG parameters. Used for testing.
    gpg_opts = ''

    def __init__(self, encrypted_file, write_lock=False):
        """Context manager that opens an encrypted file.

        Use the read() and write() methods in the subordinate "with"
        block to manipulate cleartext content. If the cleartext content
        has been replaced, the encrypted file is updated.

        `write_lock` must be set True if a modification of the file is
        intended.
        """
        self.encrypted_file = encrypted_file
        self.write_lock = write_lock

    def __enter__(self):
        self._lock()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.lockfd.close()

    @property
    def cleartext(self):
        if self._cleartext is None:
            return NEW_FILE_TEMPLATE
        return self._cleartext

    @cleartext.setter
    def cleartext(self, value):
        self._cleartext = value
        self.config = ConfigParser.ConfigParser()
        self.config.readfp(StringIO.StringIO(value))

    def read(self):
        if self._cleartext is None:
            if os.stat(self.encrypted_file).st_size:
                self._decrypt()
        return self.cleartext

    def write(self, cleartext):
        """Replace encrypted file with new content."""
        if not self.write_lock:
            raise RuntimeError('write() needs a write lock')
        self.cleartext = cleartext
        self._encrypt()

    def _lock(self):
        self.lockfd = open(
            self.encrypted_file, self.write_lock and 'a+' or 'r+')
        if self.write_lock:
            fcntl.lockf(self.lockfd, fcntl.LOCK_EX)
        else:
            fcntl.lockf(self.lockfd, fcntl.LOCK_SH)

    def _decrypt(self):
        self.cleartext = subprocess.check_output(
            ['gpg --quiet --no-tty {} --decrypt {}'.format
                (self.gpg_opts, self.encrypted_file)],
            shell=True)

    def _encrypt(self):
        recipients = self.config.get('batou', 'members').split(',')
        recipients = [x.strip() for x in recipients]
        recipients = filter(None, recipients)
        if not recipients:
            raise ValueError("Need at least one recipient.")
        recipients = ' '.join(['-r {}'.format(r.strip()) for r in recipients])
        os.rename(self.encrypted_file, self.encrypted_file+'.old')
        try:
            gpg = subprocess.Popen(
                ['gpg --quiet --no-tty {} --batch --encrypt {} -o {}'.format(
                    self.gpg_opts, recipients, self.encrypted_file)],
                stdin=subprocess.PIPE,
                shell=True)
            gpg.communicate(self.cleartext)
            if gpg.returncode != 0:
                raise RuntimeError('GPG returned non-zero exit code.')
        except Exception:
            os.rename(self.encrypted_file+'.old', self.encrypted_file)
            raise
        else:
            os.unlink(self.encrypted_file+'.old')



from batou.component import Component
from batou.lib.python import VirtualEnv, Package
from batou.lib.file import SyncDirectory
from batou.lib.supervisor import Program
from batou.utils import Address


class Django(Component):

    def configure(self):
        self.address = Address(self.host.fqdn, '8080')
        self += VirtualEnv('2.7')
        self += Package('django',
                        version='1.5.4')
        self += SyncDirectory('mysite', source='mysite')
        self += Program(
            'django',
            command='bin/python',
            args=self.expand('mysite/manage.py runserver '
                             ' {{component.address.listen}}'))



# -*- coding: utf-8 -*-

from batou.lib.file import Content, Mode, Symlink, File
from batou.lib.file import ensure_path_nonexistent
from batou.lib.file import Presence, Directory, FileComponent
from mock import Mock, patch
from stat import S_IMODE
import getpass
import os
import pytest


def test_ensure_path_nonexistent_removes_normal_file(tmpdir):
    open('asdf', 'w').close()
    assert os.path.exists('asdf')
    ensure_path_nonexistent('asdf')
    assert not os.path.exists('asdf')


def test_ensure_path_nonexistent_removes_normal_symlink(tmpdir):
    os.chdir(str(tmpdir))
    open('target', 'w').close()
    os.symlink('target', 'link')
    assert os.path.exists('link')
    assert os.path.exists('target')

    ensure_path_nonexistent('link')

    assert not os.path.exists('link')
    assert os.path.exists('target')


def test_ensure_path_nonexistent_removes_broken_symlink(tmpdir):
    os.chdir(str(tmpdir))
    os.symlink('target', 'link')
    assert os.path.islink('link')
    assert not os.path.exists('target')
    ensure_path_nonexistent('link')
    assert not os.path.exists('link')
    assert not os.path.exists('target')


def test_ensure_path_nonexistent_removes_directory(tmpdir):
    os.chdir(str(tmpdir))
    os.mkdir('dir')
    assert os.path.exists('dir')
    ensure_path_nonexistent('dir')
    assert not os.path.exists('dir')


def test_ensure_path_does_not_fail_on_nonexisting_path():
    assert not os.path.exists('missing')
    ensure_path_nonexistent('missing')
    assert not os.path.exists('missing')


def test_presence_creates_nonexisting_file(root):
    p = Presence('path')
    root.component += p
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == ''


def test_presence_leaves_existing_file_with_content_intact(root):
    p = Presence('path')
    root.component += p
    with open(p.path, 'w') as f:
        f.write('Hello there!')
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'Hello there!'


def test_presence_creates_directories_if_configured(root):
    p = Presence('directory/file', leading=True)
    root.component += p
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == ''


def test_presence_doesnt_create_directories_by_default(root):
    root.component += Presence('directory/file')
    with pytest.raises(IOError):
        root.component.deploy()


def test_presence_removes_conflicting_symlinks(root):
    p = Presence('link')
    root.component += p
    os.symlink('target', p.path)
    assert os.path.islink(p.path)
    with pytest.raises(IOError):
        open(p.path)
    root.component.deploy()
    assert not os.path.islink(p.path)
    with open(p.path) as f:
        assert f.read() == ''


def test_presence_removes_conflicting_directories(root):
    p = Presence('dir')
    root.component += p
    os.mkdir(p.path)
    root.component.deploy()
    assert not os.path.isdir(p.path)
    with open(p.path) as f:
        assert f.read() == ''


def test_directory_creates_directory(root):
    path = 'dir'
    d = Directory(path)
    root.component += d
    assert not os.path.isdir(d.path)
    root.component.deploy()
    assert os.path.isdir(d.path)


def test_directory_creates_leading_directories_if_configured(root):
    path = 'directory/path'
    d = Directory(path, leading=True)
    root.component += d
    root.component.deploy()
    assert os.path.isdir(d.path)


def test_directory_doesnt_create_leading_directories_by_default(root):
    path = 'directory/path'
    root.component += Directory(path)
    with pytest.raises(OSError):
        root.component.deploy()


def test_filecomponent_baseclass_carries_path(root):
    path = 'path'
    p = FileComponent(path)
    root.component += p
    assert p.path.endswith(path)
    assert p.original_path == path


def test_content_passed_by_string(root):
    path = 'path'
    p = Content(path, content='asdf')
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'asdf'


def test_content_passed_by_string_template(root):
    path = 'path'
    root.component.foobar = 'asdf'
    p = Content(path,
                content='{{component.foobar}}')
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'asdf'


def test_content_with_unicode_has_ascii_default(root):
    path = 'path'
    root.component.foobar = u'äsdf'
    p = File(path,
             content=u'örks {{component.foobar}}')

    with pytest.raises(UnicodeEncodeError):
        root.component |= p

    p = File(path,
             content=u'örks {{component.foobar}}',
             encoding='utf-8')
    root.component += p
    root.component.deploy()
    with open(p.path) as f:
        result = f.read().decode('utf-8')
        # XXX pytest reporting breaks if this fails. :(
        assert result == u'örks äsdf'


def test_content_passed_by_string_notemplate(root):
    path = 'path'
    root.component.foobar = 'asdf'
    p = Content(path,
                content='{{component.foobar}}',
                is_template=False)
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == '{{component.foobar}}'


def test_content_passed_by_file(root):
    source = 'source'
    with open(source, 'w') as f:
        f.write('content from source file')
    path = 'path'
    p = Content(path, source=source)
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'content from source file'


def test_content_passed_by_file_template(root):
    source = 'source'
    with open(source, 'w') as f:
        f.write('{{component.foobar}}')
    path = 'path'
    root.component.foobar = 'asdf'
    p = Content(path, source=source, is_template=True)
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'asdf'


def test_content_from_file_as_template_guessed(root):
    path = 'path'
    with open(path, 'w') as f:
        f.write('content from source file {{component.foo}}')
    p = File(path)
    root.component.foo = 'bar'
    root.component += p
    assert p.source == root.defdir + '/path'
    root.component.deploy()
    with open(p.path) as f:
        # Actually, as source and target are the same: nothing happened.
        assert f.read() == 'content from source file bar'


def test_content_passed_by_file_using_path_as_default(root):
    path = 'path'
    with open(path, 'w') as f:
        f.write('content from source file')
    p = Content(path)
    root.component += p
    with open(p.path, 'w') as f:
        # Ensure file exists
        pass
    assert p.source == root.defdir + '/path'
    root.component.deploy()
    with open(p.path) as f:
        # Actually, as source and target are the same: nothing happened.
        assert f.read() == 'content from source file'


def test_content_template_with_explicit_context(root):
    path = 'path'
    context = Mock()
    context.foobar = 'asdf'
    p = Content(path,
                content='{{component.foobar}}',
                is_template=True,
                template_context=context)
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'asdf'


def test_content_relative_source_path_computed_wrt_definition_dir(root):
    path = 'path'
    source = 'source'
    with open(source, 'w') as f:
        f.write('asdf')
    p = Content(path, source=source)
    root.component += p
    with open(p.path, 'w') as f:
        # The content component assumes there's a file already in place. So
        # we need to create it.
        pass
    root.component.deploy()
    with open(p.path) as f:
        assert f.read() == 'asdf'


def test_content_only_required_changes_touch_file(root):
    path = 'path'
    p = Content(path, content='asdf')
    root.component += p
    # Lets start with an existing file that has the wrong content:
    with open(p.path, 'w') as f:
        f.write('bsdf')
    os.utime(p.path, (0, 0))
    # Modify mtime so we can prove it's been changed.
    root.component.deploy()
    stat = os.stat(p.path)
    assert stat.st_mtime != 0

    os.utime(p.path, (0, 0))
    root.component.deploy()
    stat = os.stat(p.path)
    assert stat.st_mtime == 0


def test_content_does_not_allow_both_content_and_source(root):
    path = 'path'
    with pytest.raises(ValueError):
        root.component += Content(path, content='asdf', source='bsdf')


def test_mode_ensures_mode_for_files(root):
    path = 'path'
    open('work/mycomponent/' + path, 'w').close()
    mode = Mode(path, mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


def test_mode_ensures_mode_for_directories(root):
    path = 'path'
    os.makedirs('work/mycomponent/path')
    mode = Mode(path, mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.stat(mode.path).st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


def test_mode_ensures_mode_for_symlinks(root):
    # XXX why doesn't skipIf work correctly?!?
    if not hasattr(os, 'lchmod'):
        pytest.skip('requires lchmod')
    # This test is only relevant on platforms that support managing the mode of
    # symlinks.
    link_to = 'link_to'
    open(link_to, 'w').close()
    os.symlink(link_to, 'work/mycomponent/path')
    mode = Mode('path', mode=0o000)
    root.component += mode
    root.component.deploy()
    assert S_IMODE(os.lstat('work/mycomponent/path').st_mode) == 0o000

    mode.mode = 0o777
    root.component.deploy()
    assert S_IMODE(os.lstat('work/mycomponent/path').st_mode) == 0o777
    assert mode.changed

    root.component.deploy()
    assert not mode.changed


def test_mode_does_not_break_on_platforms_without_lchmod(root):
    if hasattr(os, 'lchmod'):
        pytest.skip('requires not having lchmod')
    # This test is only relevant on platforms without lchmod. We basically
    # ensure that deploying the component doesn't break but it's a noop
    # anyway.
    path = 'path'
    link_to = 'link_to'
    open(link_to, 'w').close()
    mode = Mode(path, mode=0o000)
    root.component += mode
    os.symlink(link_to, mode.path)
    root.component.deploy()


def test_symlink_creates_new_link(root):
    link = 'path'
    link_to = 'link_to'
    symlink = Symlink(link, source=link_to)
    root.component += symlink
    root.component.deploy()
    assert os.readlink('work/mycomponent/' + link) == symlink.source


def test_symlink_updates_existing_link(root):
    link = 'path'
    link_to = 'link_to'
    # Create an initial link
    symlink = Symlink(link, source=link_to)
    root.component += symlink
    root.component.deploy()
    # Update link with other target
    link_to2 = 'link_to2'
    symlink.source = link_to2
    root.component.deploy()
    assert os.readlink('work/mycomponent/' + link) == link_to2


def test_file_creates_subcomponent_for_presence(root):
    path = 'path'
    file = File(path)
    assert file.ensure == 'file'
    root.component += file
    assert isinstance(file.sub_components[0], Presence)


def test_file_creates_subcomponent_for_directory(root):
    file = File('dir', ensure='directory')
    root.component += file
    root.component.deploy()
    assert isinstance(file.sub_components[0], Directory)


def test_file_creates_subcomponent_for_symlink(root):
    file = File(
        'link', ensure='symlink', link_to='target')
    root.component += file
    root.component.deploy()
    assert isinstance(file.sub_components[0], Symlink)


def test_file_prohibits_unknown_ensure_parameter(root):
    with pytest.raises(ValueError):
        root.component += File('file', ensure='pipe')


@pytest.mark.slow
def test_directory_copies_all_files(root):
    os.mkdir('source')
    open('source/one', 'w').close()
    open('source/two', 'w').close()
    root.component += Directory('target', source='source')
    root.component.deploy()
    assert sorted(os.listdir('work/mycomponent/target')) == ['one', 'two']


@pytest.mark.slow
def test_directory_does_not_copy_excluded_files(root):
    os.mkdir('source')
    open('source/one', 'w').close()
    open('source/two', 'w').close()
    p = Directory(
        'target',
        source='source',
        exclude=('two',))
    root.component += p
    root.component.deploy()
    assert len(os.listdir('work/mycomponent/target')) == 1


@patch('os.chown')
def test_owner_lazy(chown, root):
    with open('asdf', 'w'):
        pass
    file = File('asdf', owner=getpass.getuser())
    root.component += file
    root.component.deploy()
    assert not os.chown.called


@patch('os.chown')
@patch('os.stat')
def test_owner_calls_chown(chown, stat, root):
    os.stat.return_value = Mock()
    os.stat.return_value.st_uid = 0
    os.stat.return_value.st_mode = 0
    file = File('asdf', owner=getpass.getuser())
    root.component += file
    root.component.deploy()
    assert os.chown.called


def test_owner_is_configurable_when_user_doesnt_exist_yet(root):
    file = File('asdf', owner='foobar')
    # This is a regression test against #12911 and ensures that we can
    # configure a file component's owner even if the owner doesn't exist yet.
    root.component += file



from batou.component import Component
from batou.lib.buildout import Buildout


class Py27(Component):

    def configure(self):
        buildout = Buildout(
            python='2.7',
            setuptools='1.4.1',
            version='2.2.1')
        self += buildout



from batou.component import Component
from batou.lib.buildout import Buildout
from batou.lib.service import Service


class Supervisor(Component):

    def configure(self):
        self.programs = self.require('programs', self.host)
        self += Buildout('supervisor',
                version='2.2',
                setuptools='0.9.8',
                python='2.7')
        self += Service('bin/supervisord',
                        pidfile='var/supervisord.pid')

    def verify(self):
        self.assert_file_is_current(
            'var/supervisord.pid',
            ['.batou.buildout.success'] + [d['path'] for d in self.programs])

    def update(self):
        out, err = self.cmd('bin/supervisorctl pid')
        try:
            int(out)
        except ValueError:
            self.cmd('bin/supervisord')
        else:
            self.cmd('bin/supervisorctl reload')
